/*
 *      Interactive disassembler (IDA).
 *      Copyright (c) 1990-98 by Ilfak Guilfanov.
 *      ALL RIGHTS RESERVED.
 *                              E-mail: ig@estar.msk.su, ig@datarescue.com
 *                              FIDO:   2:5020/209
 *
 */

#include "i51.hpp"

#define AT   COLSTR("@", SCOLOR_SYMBOL)
#define PLUS COLSTR("+", SCOLOR_SYMBOL)

static const char *phrases[] =
{
  AT COLSTR("R0", SCOLOR_REG),
  AT COLSTR("R1", SCOLOR_REG),
  AT COLSTR("DPTR", SCOLOR_REG),
  AT COLSTR("A", SCOLOR_REG) PLUS COLSTR("DPTR", SCOLOR_REG),
  AT COLSTR("A", SCOLOR_REG) PLUS COLSTR("PC", SCOLOR_REG)
};

//----------------------------------------------------------------------
inline void near OutReg(int rgnum)
{
  out_register(ph.regNames[rgnum]);
}

//----------------------------------------------------------------------
// generate the text representation of an operand

int outop(op_t &x)
{
  ulong v;
  const char *ptr;
  int dir, bit;
  switch ( x.type )
  {
    case o_reg:
      OutReg(x.reg);
      break;

    case o_phrase:
      if ( x.phrase == fRi )
      {
        out_symbol('@');
        OutReg(x.indreg);
        break;
      }
      out_colored_register_line(phrases[x.phrase]);
      break;

    case o_displ:
      out_symbol('@');
      OutReg(x.reg);
      out_symbol('+');
      OutValue(x, OOF_ADDR | OOFS_IFSIGN | OOFW_16);
      break;

    case o_imm:
      out_symbol('#');
      if ( cmd.auxpref & aux_0ext ) out_symbol('0');
      if ( cmd.auxpref & aux_1ext ) out_symbol('1');
      OutValue(x, OOFS_IFSIGN | OOFW_IMM);
      break;

    case o_mem:
    case o_near:
      v = (x.type==o_mem)
        ? map_addr(x.addr)
        : toEA(cmd.cs,x.addr);
      ptr = get_name_expr(cmd.ea+x.offb, x.n, v, x.addr);
      if( ptr == NULL )
      {
        int nbit;
        if ( cmd.itype == I51_ecall || cmd.itype == I51_ejmp )
          nbit = OOFW_32;
        else
          nbit = OOFW_16;
        OutValue(x, OOF_ADDR | OOF_NUMBER | OOFS_NOSIGN | nbit);
        QueueMark(Q_noName,cmd.ea);
        break;
      }

      // we want to output SFR register names always in COLOR_REG,
      // so remove the color tags and output it manually:

      if ( x.type == o_mem && x.addr >= 0x80 )
      {
        tag_remove(ptr, (char *)ptr, 0);
        out_register(ptr);
        break;
      }
      OutLine(ptr);
      break;

    case o_void:
      return 0;

    case o_bit251:
      if ( x.b251_bitneg ) out_symbol('/');
      dir = x.addr;
      bit = x.b251_bit;
      goto OUTBIT;

    case o_bitnot:
      out_symbol('/');
    case o_bit:
      dir = (x.reg & 0xF8);
      bit = x.reg & 7;
      if ( (dir & 0x80) == 0 ) dir = dir/8 + 0x20;
OUTBIT:
      if(ash.uflag & UAS_PBIT)
      {
        predefined_t *predef = GetPredefined(ibits, dir, bit);
        if ( predef != NULL )
        {
          out_line(predef->name, COLOR_REG);
          break;
        }
      }
      {
        v = map_addr(dir);
        char *rname = get_name_expr(cmd.ea+x.offb, x.n, v, dir);
        if ( rname != NULL && strchr(rname, '+') == NULL )
        {

      // we want to output the bit names always in COLOR_REG,
      // so remove the color tags and output it manually:

          if ( dir < 0x80 )
          {
            OutLine(rname);
          }
          else
          {
            tag_remove(rname, rname, 0);
            out_register(rname);
          }
        }
        else
        {
          out_long(dir, 16);
        }
        out_symbol(ash.uflag & UAS_NOBIT ? '_' : '.');
        out_symbol('0'+bit);
      }
      break;

     default:
       warning("out: %lx: bad optype",cmd.ea,x.type);
       break;
  }
  return 1;
}

//----------------------------------------------------------------------
// generate a text representation of an instruction
// the information about the instruction is in the 'cmd' structure

void out(void)
{
  char buf[MAXSTR];

  u_line = buf;                         // setup the output pointer
  OutMnem();                            // output instruction mnemonics

  out_one_operand(0);                   // output the first operand

  if ( cmd.Op2.type != o_void)
  {
    out_symbol(',');
    OutChar(' ');
    out_one_operand(1);                 // output the second operand
  }

  if ( cmd.Op3.type != o_void)
  {
    out_symbol(',');
    OutChar(' ');
    out_one_operand(2);                 // output the third operand
  }


  // output a character representation of the immediate values
  // embedded in the instruction as comments

  if ( isVoid(cmd.ea,uFlag,0) ) OutImmChar(cmd.Op1);
  if ( isVoid(cmd.ea,uFlag,1) ) OutImmChar(cmd.Op2);
  if ( isVoid(cmd.ea,uFlag,2) ) OutImmChar(cmd.Op3);

  while ( u_line[-1] == ' ' ) u_line--; // trim the trailing spaces
  *u_line = '\0';                       // terminate the output string
  gl_comm = 1;                          // ask to attach a possible user-
                                        // defined comment to it
  MakeLine(buf);                        // pass the generated line to the
                                        // kernel
}

//--------------------------------------------------------------------------
// generate start of the disassembly

void header(void)
{
  gen_cmt_line("Processor:        %s", inf.procName);
  gen_cmt_line("Target assembler: %s", ash.name);
  if ( ash.header != NULL )
    for ( const char **ptr=ash.header; *ptr != NULL; ptr++ ) MakeLine(*ptr,0);
}

//--------------------------------------------------------------------------
// generate start of a segment

void segstart(ea_t ea)
{
  char buf[MAXSTR];
  segment_t *Sarea = getseg(ea);
    if ( ash.uflag & UAS_SECT ) {
      if ( Sarea->type == SEG_IMEM ) MakeLine(".RSECT");
      else {
        snprintf(buf, sizeof(buf), COLSTR("%s: .section", SCOLOR_ASMDIR),
                                                       get_segm_name(Sarea));
        MakeLine(buf,0);
      }
    } else {
      if(ash.uflag & UAS_NOSEG)
          snprintf(buf, sizeof(buf), COLSTR("%s.segment %s", SCOLOR_AUTOCMT),
                   ash.cmnt, get_segm_name(Sarea));
      else snprintf(buf, sizeof(buf), COLSTR("segment %s",SCOLOR_ASMDIR),
                                                       get_segm_name(Sarea));
      MakeLine(buf);
      if(ash.uflag & UAS_SELSG) MakeLine(get_segm_name(Sarea));
      if(ash.uflag & UAS_CDSEG)
        MakeLine((Sarea->type == SEG_IMEM) ?
              COLSTR("DSEG", SCOLOR_ASMDIR) : COLSTR("CSEG", SCOLOR_ASMDIR));
              // XSEG - eXternal memory
    }
    if ( inf.s_org ) {
      ulong org = ea - get_segm_base(Sarea);
      if( org != 0 ) gen_cmt_line("%s %s", ash.origin, btoa(org));
    }
}

//--------------------------------------------------------------------------
// generate end of the disassembly

void footer(void)
{
  char buf[MAXSTR];
  if ( ash.end != NULL)
  {
    MakeNull();
    register char *p = tag_addstr(buf, COLOR_ASMDIR, ash.end);
    const char *start = get_colored_name(BADADDR, inf.beginEA);
    if ( start != NULL)
    {
      *p++ = ' ';
      if(ash.uflag & UAS_NOENS) p = stpcpy(p, ash.cmnt);
      strcpy(p, start);
    }
    MakeLine(buf);
  }
  else
  {
    gen_cmt_line("end of file");
  }
}

//--------------------------------------------------------------------------
// output one "equ" directive

static void near do_out_equ(char *name,const char *equ,uchar off)
{
  char buf[MAXSTR];
  register char *p;
  gl_name = 0;
  if (ash.uflag & UAS_PSAM) {
    p = tag_addstr(buf, COLOR_KEYWORD, equ);
    *p++ = ' ';
    p = stpcpy(p, name);
    p = tag_addchr(p, COLOR_SYMBOL, ',');
  } else {
    p = stpcpy(buf, name);
    if(ash.uflag & UAS_EQCLN) p = tag_addchr(p, COLOR_SYMBOL, ':');
    *p++ = ' ';
    p = tag_addstr(p, COLOR_KEYWORD, equ);
    *p++ = ' ';
  }
  tag_addstr(p, COLOR_NUMBER, btoa(off));
  MakeLine(buf,0);
}

//--------------------------------------------------------------------------
// output "equ" directive(s) if necessary

static int near out_equ(ea_t ea)
{
  segment_t *s = getseg(ea);
  if ( s != NULL && s->type == SEG_IMEM && ash.a_equ != NULL)
  {
    char *name = get_name(BADADDR, ea);
    if ( name != NULL
      && ((ash.uflag & UAS_PBYTNODEF) == 0 || !IsPredefined(name)) )
    {
      char buf[MAXSTR];
      char *ptr = stpcpy(buf, get_colored_name(BADADDR, ea));
      uchar off = uchar(ea - get_segm_base(s));
      do_out_equ(buf, ash.a_equ, off);
      if ( (ash.uflag & UAS_AUBIT) == 0 && (off & 0xF8) == off )
      {
        ptr = tag_on(ptr, COLOR_SYMBOL);
        *ptr++ = ash.uflag & UAS_NOBIT ? '_' : '.';
        *ptr++ = '0';
        *tag_off(ptr, COLOR_SYMBOL) = '\0';
        for( ; ptr[-1] < '8'; off++, (ptr[-1])++)
          do_out_equ(buf, ash.a_equ, off);
        MakeNull();
      }
    }
    else
    {
      gl_name = 0;
      MakeLine("");
    }
    return 1;
  }
  if ( ash.uflag & UAS_NODS )
  {
    if ( !hasValue(getFlags(ea)) && s->type == SEG_CODE )
    {
      char buf[MAXSTR];
      ulong org = ea - get_segm_base(s) + getSize(ea);
      sprintf(buf, COLSTR("%s %s", SCOLOR_ASMDIR), ash.origin, btoa(org));
      MakeLine(buf);
      return 1;
    }
  }
  return 0;
}

//--------------------------------------------------------------------------
// generate a data representation
// usually all the job is handled by the kernel's standard procedure,
// intel_data()
// But 8051 has its own quirks (namely, "equ" directives) and intel_data()
// can't handle them. So we output "equ" ourselves and pass everything
// else to intel_data()

void i51_data(ea_t ea)
{
  gl_name = 1;

  // the kernel's standard routine which outputs the data knows nothing
  // about "equ" directives. So we do the following:
  //    - try to output an "equ" directive
  //    - if we succeed, then ok
  //    - otherwise let the standard data output routine, intel_data()
  //        do all the job

  if ( !out_equ(ea) )
    intel_data(ea);
}
